
package rcpsp_choco_enumeration;


import java.io.*;
import org.chocosolver.solver.Model;
import org.chocosolver.solver.*;
import org.chocosolver.solver.search.strategy.Search;
import org.chocosolver.solver.search.strategy.selectors.values.IntDomainMin;
import org.chocosolver.solver.search.strategy.selectors.variables.FirstFail;
import org.chocosolver.solver.search.strategy.strategy.IntStrategy;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.util.ESat;




public class RCPSP_Choco_Enumeration {

    public static void main(String[] args) {
        int n = 7;
        int Dmax = 22;
        int Qdispo = 4;
        int les_durees[] = new int [100] ;
        int les_conso[] = new int [100];
        int pred[][] = new int [100][100]; 
        
        
        int duree_max = 0;

        for (int i=0; i<n; i++){
            for (int j=0; j<n; j++){
             pred[i][j] = 0;
            }    
      }
            
        try{
            File  source;     
            source = new File("./instance/essai_5.txt");  
            FileReader Input=new FileReader(source);
            StreamTokenizer   Lecteur;
            Lecteur=new StreamTokenizer (Input);
            Lecteur.parseNumbers();

            Lecteur.nextToken();
            int nombre=0;
            if (Lecteur.ttype==Lecteur.TT_NUMBER){ 
                    nombre=(int) Lecteur.nval; // nombre de sommets
            }

            for (int i=1; i<=n; i++){
                Lecteur.nextToken();
                int sommet=0;
                if (Lecteur.ttype==Lecteur.TT_NUMBER){sommet=(int) Lecteur.nval ;}
                sommet--; 
                Lecteur.nextToken();// suppression du chiffre inutile
                Lecteur.nextToken();// nb successeur

                int nb_succ=0;
                if (Lecteur.ttype==Lecteur.TT_NUMBER){ nb_succ=(int) Lecteur.nval;}
                if (nb_succ!=0) Lecteur.nextToken();
                if ( (i==n) &&(nb_succ==0) )
                    Lecteur.nextToken();
                // parcours des successeurs
                for (int j=1; j<=nb_succ; j++){
                    int succ=0;
                    if (Lecteur.ttype==Lecteur.TT_NUMBER){succ=(int) Lecteur.nval;}
                    succ--;
                    pred[sommet][succ]=1;
                    if (j!=nb_succ) Lecteur.nextToken();
                }
            }
            

            // bloc 2 : lecture des durées et des consommations
            for (int i=1; i<=n; i++){
                int sommet=0;
                if (Lecteur.ttype==Lecteur.TT_NUMBER)
                        {sommet=(int) Lecteur.nval;}
                sommet--;
                // suppression du chiffre inutile
                Lecteur.nextToken();
                Lecteur.nextToken();
                // lecture conso
                int conso=0;
                if (Lecteur.ttype==Lecteur.TT_NUMBER){
                    conso=(int) Lecteur.nval;
                    les_conso[sommet]=conso;}

                Lecteur.nextToken();
                // lecture duree
                int duree=0;
                if (Lecteur.ttype==Lecteur.TT_NUMBER){
                    duree=(int) Lecteur.nval;
                    les_durees[sommet]=duree;
                }
                Lecteur.nextToken();
            }

            // lecture duree
            int qt_dispo=0;
            if (Lecteur.ttype==Lecteur.TT_NUMBER){
                qt_dispo=(int) Lecteur.nval;
                Qdispo = qt_dispo;
            }

            for (int i=1; i<=n; i++){
                if (les_durees[i]>duree_max){duree_max = les_durees[i];}
            }
           
            Input.close();
            }catch (Exception e) {
                  System.out.println(e.getMessage());
            };  
            
       // -------  fin lecture fichier -----------//
            
            Model mon_modele = new Model("RCPSP");          
            
            
            // contrainte 1  
           IntVar[] St = new IntVar[n];
           for (int i=1; i<n-1; i++){
                St[i] = mon_modele.intVar("St_"+i, 0, Dmax-les_durees[i]);
           }

           // contraintes 1.1. et 1.2
           St[n-1] = mon_modele.intVar("St_"+ (n-1), duree_max, Dmax);
           St[0] = mon_modele.intVar("St_0", 0);


           // contrainte 2
           IntVar[] Ft = new IntVar[n];
           for (int i=0; i<n; i++){
                Ft[i] = mon_modele.intVar("Ft_"+i,les_durees[i], Dmax);
            }

         
           // contrainte 3 
           IntVar[][] qt = new IntVar[n][n];
           for (int i=0; i<n; i++){
                for (int j=1; j<n; j++){
                   int min = les_conso[i];  
                   if (i==0){min=Qdispo;}
                   if (j!=i){
                        int valeur_ref = les_conso[j];
                        if (j==n-1){valeur_ref = Qdispo;}
                        if (valeur_ref<min) {min=valeur_ref;}           
                            qt[i][j] = mon_modele.intVar("qt_"+i+"_"+j, 0, min);
                    }
                }
            }  
           
           // contraainte 3.1
            for (int i=0; i<n; i++){
                for (int j=0; j<n; j++){            
                    if (pred[i][j]==1){
                        qt[i][j] = mon_modele.intVar("qt_"+i+"_"+j, 0);
                    }
                }
            }
           
            // contrainte 3.2
           for (int i=0; i<n; i++){
               qt[i][i] = mon_modele.intVar("qt_"+i+"_"+0, 0);
           }
                      
           // contrainte 3.3
           for (int i=1; i<n; i++){
               qt[i][0] = mon_modele.intVar("qt_"+i+"_"+0, 0);
           }

           // contrainte 4
           IntVar OBJ = mon_modele.intVar("Cmax", 0, Dmax);      
           
           // contrainte 5
           for (int i=0; i<n; i++){
             mon_modele.arithm(Ft[i],"=",St[i],"+",les_durees[i]).post();
           }

           // // contrainte 6: cas de 0
           mon_modele.sum(qt[0],"=",Qdispo).post();

           // contrainte 7
           int[] coeffs = new int[n];
           for (int j=0; j<n ; j++){
               coeffs[j] = 1;  
           }
           for (int i=1; i<n-1; i++){
                IntVar[] colonne = new IntVar[n];            
                for (int j=0; j<n ; j++){
                   colonne[j] = qt[j][i];  
                 }
                 mon_modele.scalar(colonne, coeffs,"=",les_conso[i]).post();
           }

           // contrainte 8
           IntVar[] colonne = new IntVar[n];            
           for (int j=0; j<n ; j++){
               colonne[j] = qt[j][n-1];  
           }
           mon_modele.scalar(colonne, coeffs,"=",Qdispo).post();     


           // contrainte 9
           for (int i=1; i<n-1; i++){
               mon_modele.sum(qt[i],"=",les_conso[i]).post();
           }


            // contrainte 10
             for (int i=0; i<n; i++){
               for (int j=0; j<n; j++) {            
                  mon_modele.ifThen(
                  mon_modele.arithm(qt[i][j],">",0),
                  mon_modele.and(mon_modele.arithm(St[j],">=",Ft[i]),mon_modele.arithm(qt[j][i],"=",0))
                  );              
               }
             }         


            // contrainte 11
           for (int i=0; i<n; i++){
               for (int j=0; j<n; j++)  {            
                 if (pred[i][j]==1) {
                   mon_modele.arithm(St[i],">=",Ft[j]).post();
                 }
               }
             }

           // contrainte 11.1
            for (int i=0; i<n; i++){
                for (int j=0; j<n; j++)  { 
                  for (int k=0; k<n; k++){ 
                    if ((pred[i][j]==1) && (pred[j][k]==1)){
                        mon_modele.arithm(qt[k][i],"=",0).post();
                        }
                    }
                }
            }

           
            // contrainte 12
            OBJ=Ft[n-1];
            

            // fonction objectif
            mon_modele.setObjective(Model.MINIMIZE, OBJ);     
            Solution s = new Solution(mon_modele);

           

     
           // Configuration du solveur
           Solver mon_solveur = mon_modele.getSolver();

            IntVar[] qt1D = new IntVar[(n-1)*(n)];
            for (int i=0; i<n-1; i++){
                for (int j=0; j<n; j++){
                    qt1D[i*n+j] = qt[i][j];   
                    
                }
            }


           Mon_Selecteur selecteur_1 = new Mon_Selecteur(); 
           choisir_Value mon_select_value_1 =  new choisir_Value();
           IntStrategy s1 = Search.intVarSearch( selecteur_1,
                                                 mon_select_value_1, 
                                                 St);

           mon_solveur.setSearch( 
                   s1,
                   Search.intVarSearch ( 
                           new FirstFail(mon_modele), 
                           new IntDomainMin(), 
                           qt1D)                
                   );

            
           mon_solveur.showDecisions();


            while (mon_solveur.solve()) {
                s.record();
                int cout = s.getIntVal(St[n-1]);
                System.out.println("    Cmax = " + cout);
              }  

              System.out.println("*****************");
              if (mon_solveur.isFeasible() == ESat.TRUE)
               {
                int cout = s.getIntVal(St[n-1]);
               System.out.println("Cmax = " + cout);
               System.out.println("Solution = ");

                for (int i=0; i<n; i++)
                {
                    System.out.print( ""+i+" : "+s.getIntVal(St[i]) + " - ");
                    System.out.println(s.getIntVal(Ft[i]) + " ");
                }

                System.out.println( "Flot  :");
                for (int i=0; i<n; i++){
                  for (int j=1; j<n; j++){                 
                    if (s.getIntVal(qt[i][j])>0)
                      System.out.println(""+i+" - "+j+" : "+ s.getIntVal(qt[i][j]));
                  }
                } 
               }
               else
                 System.out.println("pas de sol.faisable");
                 
     
        
    }
    
}
